#include "mainNode/MainNode.h" 

#include "exceptions/Exceptions.h"

#include "PlatoonProtocolLib/NetworkInfo.h"

#include <pluginlib/class_list_macros.h>

#include <iostream>

#include <fstream>
#include <sstream>
#include <vector>
#include <iterator>

PLUGINLIB_EXPORT_CLASS(car::MainNode, nodelet::Nodelet);

namespace car
{
  
platoonProtocol::NetworkInfo readNetworkInfo() {
  /** This reads ~/.CarConfig/platoon.config */
  
  // open config file
  std::string userHome = getenv("HOME");
  std::ifstream configFile;
  configFile.open(userHome + "/CarConfig/platoon.config", std::ifstream::in);
  if (!configFile.is_open()) { throw FileNotFound(); }

  // desired parameters
  platoonProtocol::VehicleId vehicleId;
  bool foundVehicleId = false;
  std::string broadCastAdress;
  bool foundBroadCastAdress = false;
    
  std::string contentLine;
  while (!configFile.eof()) {
    std::getline(configFile, contentLine);
     
    // split this line
    std::istringstream iss(contentLine);
    std::vector<std::string> words {std::istream_iterator<std::string>(iss), {} };
    
    if ( words.size() > 1) { // this line contains parameter[at(0)] and value[at(1)]
      if ( words.at(0) == "VehicleId:" ) {
        vehicleId = std::stoi(words.at(1));
        foundVehicleId = true;
      } else if ( words.at(0) == "BroadCast:" ) {
        broadCastAdress = words.at(1);
        foundBroadCastAdress = true;
      }
    }
    
  }
  
  if ( !(foundVehicleId && foundBroadCastAdress) ) { throw MissingParameters(); }
  configFile.close();

  return platoonProtocol::NetworkInfo{vehicleId, broadCastAdress};

}
  
MainNode::MainNode(ros::NodeHandle & nh, std::string & name) :
    nh(nh)  
    , name(name)
    , c2cNet()
    , c2c(platoonProtocol::Vehicle::Role::FOLLOWER, c2cNet, readNetworkInfo())
    , pcNet()
    , pc(pc2car::CommandReceiver::create(pcNet))
    , egoMotion(nh)
    , platoonController(c2c, pc, egoMotion)
    , platoonControllerThread([this] { platoonController.run(); })
    , cruiseController(c2c, egoMotion, platoonController)
    , cruiseControllerThread([this] { cruiseController.run(); })
{
  egoMotion.cruiseControllerNotify = cruiseControllerNotify;
  egoMotion.platoonControllerNotify = platoonControllerNotify;
  platoonController.cruiseControllerNotify = cruiseControllerNotify;
  pc->receiveCommands([this] (auto commandCode) {platoonControllerThread.notify();} );
  c2c.setCallback( [this] {platoonControllerThread.notify();} );
}

MainNode::MainNode() : 
    nh()  
    , name()
    , c2cNet()
    , c2c(platoonProtocol::Vehicle::Role::FOLLOWER, c2cNet, readNetworkInfo())
    , pcNet()
    , pc(pc2car::CommandReceiver::create(pcNet))
    , egoMotion(nh)
    , platoonController(c2c, pc, egoMotion)
    , platoonControllerThread([this] { platoonController.run(); })
    , cruiseController(c2c, egoMotion, platoonController)
    , cruiseControllerThread([this] { cruiseController.run(); })
{
  egoMotion.cruiseControllerNotify = cruiseControllerNotify;
  egoMotion.platoonControllerNotify = platoonControllerNotify;
  platoonController.cruiseControllerNotify = cruiseControllerNotify;
  pc->receiveCommands([this] (auto commandCode) {platoonControllerThread.notify();} );
  c2c.setCallback( [this] {platoonControllerThread.notify();} );
}

MainNode::~MainNode()
{}

void MainNode::onInit()
{
    NODELET_INFO("MainNode::onInit -- START");
    NODELET_INFO("MainNode::onInit -- END");
}

}
